import datetime
import json
from typing import *
import dataclasses
from collections import abc


class JsonEncoder(json.JSONEncoder):
    def default(self, o: "Any") -> "Any":
        if isinstance(o, (datetime.date, datetime.datetime)):
            return o.isoformat()
        elif hasattr(o, "to_dict"):
            return o.to_dict()
        return json.JSONEncoder.default(self, o)

def to_dict(obj, item_class=dict, collection_class=list, date_style="python"):
    # print(f"Obj is {obj}")
    if isinstance(obj, abc.Mapping):
        # print(f"Casting as Mapping with class {item_class}")
        return item_class((k, to_dict(v, item_class, collection_class, date_style)) for k, v in obj.items())
    elif dataclasses.is_dataclass(obj):
        # print("Converting Dataclass")
        return to_dict(item_class(obj), item_class, collection_class, date_style)
    elif isinstance(obj, abc.Iterable) and not isinstance(obj, (str, bytes)):
        # print(f"Casting as Collection with {collection_class}")
        return collection_class(to_dict(i, item_class, collection_class, date_style) for i in obj)
    elif date_style == "mongo" and isinstance(obj, datetime.date):
        return datetime.datetime.combine(obj, datetime.datetime.min.time())
    elif date_style == "json" and isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat()
    else:
        # print("No cast - passing through")
        return obj
    
async def ato_dict(obj, item_class=dict, collection_class=list, date_style="python"):
    if isinstance(obj, abc.AsyncIterable):
        obj = [o async for o in obj]
    return to_dict(obj, item_class, collection_class, date_style)


def resolve_list(item, key):
    item_list = resolve(item, key)

    if item_list is None:
        return []

    if isinstance(item_list, list):
        return item_list

    return [item_list]

def resolve_attribute(obj, key):
    if isinstance(obj, abc.Mapping):
        obj = obj[key]
    else:
        obj = getattr(obj, key)
    if callable(obj):
        obj = obj()
    return obj


def resolve(item, key):
    if key is None:
        return item
    accessors = key.split(".")
    try:
        for accessor in accessors:
            if isinstance(item, abc.Sequence) and accessor.isdigit():
                item = item[int(accessor)]
            elif isinstance(item, abc.Sequence) and not accessor.isdigit():
                item = [resolve_attribute(i, accessor) for i in item]
            else:
                item = resolve_attribute(item, accessor)
    except Exception as e:
        return None
    return item

class DataConversion():
    def to_mongo(self):
        """Convert object to a Python dictionary with datetime.dates converted to datetime.datetimes for MongoDB compatibility"""
        return to_dict(self, date_style="mongo")

    def to_pandas(self):
        """Convert object to Pandas Series"""
        import pandas as pd

        return pd.Series(to_dict(self))

    def to_json(self, *args, **kwargs):
        """Convert object to a JSON string"""
        return json.dumps(to_dict(self, date_style="json"), *args, **kwargs)

    def to_dict(self, item_class=dict, collection_class=list, date_style="python"):
        """Convert object to simple Python dictionary"""
        return to_dict(self, item_class=item_class, collection_class=collection_class, date_style=date_style)


import dataclasses as dc
from .dataclass_gen import generate_dc_code
from yankee.data import Row
from pathlib import Path

import pytest

@dc.dataclass
class Example(Row):
    a: int = None
    b: "List" = dc.field(default_factory=list)

    def example_function(self):
        pass
@pytest.mark.skip("WIP")
def test_generate_dc_code():
    source = generate_dc_code(Example)
    with (Path(__file__).parent / "example.py").open("w") as f:
        f.write(source)
    assert True


from toolz.functoolz import compose
from yankee.util import inflect
from yankee import settings
from .accessor import python_accessor

class DefaultMeta:
    accessor_function = python_accessor
    infer_keys = True
    output_style = "python"

class Deserializer(object):
    Meta = DefaultMeta


    def __init__(self, data_key=None, many=False, required=False):
        self.data_key = data_key
        self.required = required
        self.many = many
        self.name = None
        self._build_meta()
        self.bind()

    def _build_meta(self):
        for c in self.__class__.mro():
            if not hasattr(c, "Meta"):
                continue
            for k in filter(lambda k: not k.startswith("_"), c.Meta.__dict__.keys()):
                if not hasattr(self.Meta, k):
                    setattr(self.Meta, k, getattr(c.Meta, k))
                
    def bind(self, name=None, parent=None, meta=None):
        self.name = name
        self.parent = parent
        # Update Meta object
        if meta:
            self.Meta = meta
        elif self.parent is not None:
            self.Meta = self.parent.Meta
        # Regenerate Accessor
        self.make_accessor()
        # Set Output Name
        if self.name is not None:
            self.output_name = inflect(self.name, style=self.Meta.output_style)
        if settings.use_model:
            self._load_func = compose(self.load_model, self.post_load, self.deserialize, self.pre_load)
        else:
            self._load_func = compose(self.post_load, self.deserialize, self.pre_load)
        return self

    def make_accessor(self):
        self.accessor = self.Meta.accessor_function(self.data_key, self.name, self.many, self.Meta)

    def load(self, obj):
        self.raw = obj
        return self._load_func(obj)

    def pre_load(self, obj):
        return obj

    def load_model(self, obj):
        return obj
    
    def deserialize(self, obj):
        return self.accessor(obj)

    def post_load(self, obj):
        return obj



import pytest
import json

from .collection import Collection

class TestCollection():
    def test_sync_iterable(self, event_loop):
        iterable = ["a", "b", "c"]
        collection = Collection(iterable)
        assert [o for o in collection] == iterable
        
    def test_async_iterable(self, event_loop):
        iterable = ["a", "b", "c"]
        async def gen():
            for i in iterable:
                yield i
        collection = Collection(gen())
        assert [o for o in collection] == iterable
        
    def test_list_conversion(self):
        iterable = ["a", "b", "c"]
        collection = Collection(iterable)
        assert collection.to_list() == iterable
        
    def test_records_conversion(self):
        iterable = [{"a": 1}, {"a": 2}]
        collection = Collection(iterable)
        assert collection.to_records() == iterable
        
    def test_json_conversion(self):
        iterable = [{"a": 1}, {"a": 2}]
        collection = Collection(iterable)
        assert json.loads(collection.to_json()) == iterable
        
    def test_pandas_conversion(self):
        iterable = [{"a": 1}, {"a": 2}]
        collection = Collection(iterable)
        assert collection.to_pandas().a.tolist() == [1, 2]
        
    

class TestCollectionAsync():
    @pytest.mark.asyncio
    async def test_sync_iterable(self, event_loop):
        iterable = ["a", "b", "c"]
        collection = Collection(iterable)
        assert [o async for o in collection] == iterable
        
    @pytest.mark.asyncio
    async def test_async_iterable(self, event_loop):
        iterable = ["a", "b", "c"]
        async def gen():
            for i in iterable:
                yield i
        collection = Collection(gen())
        assert [o async for o in collection] == iterable
        
    @pytest.mark.asyncio
    async def test_list_conversion(self):
        iterable = ["a", "b", "c"]
        collection = Collection(iterable)
        assert await collection.ato_list() == iterable
        
    @pytest.mark.asyncio
    async def test_records_conversion(self):
        iterable = [{"a": 1}, {"a": 2}]
        collection = Collection(iterable)
        assert await collection.ato_records() == iterable
        
    @pytest.mark.asyncio
    async def test_json_conversion(self):
        iterable = [{"a": 1}, {"a": 2}]
        collection = Collection(iterable)
        assert json.loads(await collection.ato_json()) == iterable
        
    @pytest.mark.asyncio
    async def test_pandas_conversion(self):
        iterable = [{"a": 1}, {"a": 2}]
        collection = Collection(iterable)
        df = await collection.ato_pandas()
        assert df.a.tolist() == [1, 2]
        
        
explode_data = [
    {"a": 1, "b": [2, 3]},
    {"a": 2, "b": [4, 5]},
]
        
class TestExplode():
    def test_can_do_basic_explode(self):
        collection = Collection(explode_data)
        exploded = collection.explode("b").to_list()
        assert len(exploded) == 4
        assert exploded == [
            {"a": 1, "b": 2},
            {"a": 1, "b": 3},
            {"a": 2, "b": 4},
            {"a": 2, "b": 5},
        ]

        
class TestExplodeAsync():
    @pytest.mark.asyncio
    async def test_can_do_basic_explode(self):
        collection = Collection(explode_data)
        exploded = await collection.explode("b").ato_list()
        assert len(exploded) == 4
        assert exploded == [
            {"a": 1, "b": 2},
            {"a": 1, "b": 3},
            {"a": 2, "b": 4},
            {"a": 2, "b": 5},
        ]

unpack_data = [
    {"a": 1, "b": {"c": 2, "d": 3}},
    {"a": 2, "b": {"c": 4, "d": 5}},
]

class TestUnpack():
    def test_can_do_basic_unpack(self):
        collection = Collection(unpack_data)
        unpacked = collection.unpack("b").to_list()
        assert len(unpacked) == 2
        assert unpacked == [
            {"a": 1, "b.c": 2, "b.d": 3},
            {"a": 2, "b.c": 4, "b.d": 5},
        ]
        
class TestUnpackAsync():
    @pytest.mark.asyncio
    async def test_can_do_basic_unpack(self):
        collection = Collection(unpack_data)
        unpacked = await collection.unpack("b").ato_list()
        assert len(unpacked) == 2
        assert unpacked == [
            {"a": 1, "b.c": 2, "b.d": 3},
            {"a": 2, "b.c": 4, "b.d": 5},
        ]

values_data = [
    {"a": 1, "b": 2, "c": 3},
    {"a": 2, "b": 4, "c": 5},
]

class TestValues():
    def test_can_do_basic_values(self):
        collection = Collection(values_data)
        values = collection.values("a", "c").to_list()
        assert len(values) == 2
        assert values == [
            {"a": 1, "c": 3},
            {"a": 2, "c": 5},
        ]
        
class TestValuesAsync():
    @pytest.mark.asyncio
    async def test_can_do_basic_values(self):
        collection = Collection(values_data)
        values = [o async for o in collection.values("a", "c")]
        assert len(values) == 2
        assert values == [
            {"a": 1, "c": 3},
            {"a": 2, "c": 5},
        ]
        
class TestValuesList():
    def test_can_do_basic_values(self):
        collection = Collection(values_data)
        values = collection.values_list("a", "c").to_list()
        assert len(values) == 2
        assert values == [
            (1, 3),
            (2, 5),
        ]
        
class TestValuesListAsync():
    @pytest.mark.asyncio
    async def test_can_do_basic_values(self):
        collection = Collection(values_data)
        values = [o async for o in collection.values_list("a", "c")]
        assert len(values) == 2
        assert values == [
            (1, 3),
            (2, 5),
        ]
    
    @pytest.mark.asyncio
    async def test_can_do_for_loops(self):
        collection = Collection(values_data)
        values = list()
        async for o in collection.values_list("a", "c"):
            values.append(o)
        assert len(values) == 2
        assert values == [
            (1, 3),
            (2, 5),
        ]